module Redify
  class Core
    class << self
      
      # Valid options:
      #   :resource - ActiveRecord model class (User)
      #   :collection - camel-cased string denoting collection name
      #   :record - a single record instance (User.find 1)
      def get_namespaced_key_name_for(options)
        key_name = get_key_name_for(options)
        namespace = Redify::Settings.namespace
        namespace ? "#{namespace}:#{key_name}" : key_name
      end

      # Returns the key name without the namespace.
      def get_key_name_for(options)
        if options[:collection]
          get_key_name_for_collection(options)
        elsif options[:record]
          get_key_name_for_record(options[:record])
        elsif options[:resource] && options[:id]
          get_key_name_for_resource_and_id(options)
        elsif options[:resource]
          get_key_name_for_resource(options[:resource])
        end
      end

      def get_key_name_for_collection(options)
        key_name = get_key_name_for_resource(options[:resource])
        if options[:collection] == "All"
          key_name << ":All"
        else
          key_name << ":Collections:#{options[:collection]}"
        end
      end

      def get_key_name_for_record(record)
        "#{record.class.name}:#{record.id}"
      end

      def get_key_name_for_resource(resource)
        "#{ActiveSupport::Inflector.pluralize(resource.to_s)}"
      end

      def get_key_name_for_resource_and_id(options)
        "#{options[:resource].to_s}:#{options[:id]}"
      end

      # Valid options:
      #   :resource - ActiveRecord model class (User)
      #   :id - record primary key
      def fetch_record(options)

        resource = options[:resource]

        return nil unless resource
        return resource.find(options[:id], :skip_redify => true) unless resource.redify_handler # Redis is unavailable
        return synchronize(:resource => resource, :id => options[:id])

      end

      # fetch first or last N entries
      def fetch_records_by_edge(resource, last = false, count = 1)
        # TODO implement COUNT
        edge_id = redify_handler.lindex resource.redify_keys[:all], (last ? -1 : 0)
        edge_id ? Redify::Core.fetch_record(:resource => resource, :id => edge_id) : nil
      end

      # Loads all records into Redis upon initialization
      def initialize_resource(resource)

        redis = resource.redify_handler
        resource_status = redis.hget(resource.redify_keys[:resource], 'status')

        if resource_status == "READY"
          puts "[Redify] Resource #{resource} is already loaded."
        elsif resource_status == 'LOADING'
          puts "[Redify] Resource #{resource} is being loaded by another process."
        else
          puts "[Redify] Initializing resource #{resource}"
          redify_all_records(resource, redis)
          puts "... finished"
        end

      end

      def redify_all_records(resource, redis)
        flush_keys(resource)
        resource_key = resource.redify_keys[:resource]
        redis.hset(resource_key, 'status', 'LOADING')
        resource.find(:all, :skip_redify => true).each do |r|
          redify_record(r)
          redis.rpush(redify_keys[:all], r.id)
        end
        # TODO load associations and scopes
        redis.hset(resource_key, 'status', 'READY')
      end

      # Turns an ActiveRecord instance into a Redis entry.
      # Updates Redis entry if it already exists.
      def redify_record(record)
        key = Redify::Core.get_namespaced_key_name_for(:record => record)
        record.class.redify_handler.hmset(key, record.attributes.to_a.flatten)
        puts "[Redify] #{record.class}\##{record.id} redified"
        return record
      end

      # Callbacks for Redified models

      def after_create(record)
        record.class.redify_handler.rpush record.class.redify_keys[:all], record.id
      end

      def after_save(record)
        redify_record record
      end

      def after_destroy(record)
        puts "after_destroy"
        puts record.inspect
      end

      # Deletes all Redis keys under current namespace
      def flush_keys(resource = nil)
        namespace = Redify::Settings.namespace
        redis = Redify::Settings.default_handler
        pattern = "#{namespace}:"
        pattern << "#{resource}" if resource
        pattern << "*"
        redis.keys(pattern).each do |key|
          redis.del(key)
        end
        # namespace is also a key by itself
        redis.del(namespace) unless resource
      end


    end
  end
end